require 'sketchup.rb'
require 'extensions.rb'

# 插件扩展信息
PLUGIN_NAME = '视图旋转工具'
PLUGIN_VERSION = '1.0.0'
PLUGIN_AUTHOR = '刘涛'
PLUGIN_COPYRIGHT = '© 2025 liutao91.cn'
PLUGIN_DESCRIPTION = '可快速旋转视图，指定视图的旋转角度。'

# 注册插件扩展，使插件信息能在扩展管理器中显示
ex = SketchupExtension.new(PLUGIN_NAME, __FILE__)
ex.description = PLUGIN_DESCRIPTION
ex.version = PLUGIN_VERSION
ex.creator = PLUGIN_AUTHOR
ex.copyright = PLUGIN_COPYRIGHT
Sketchup.register_extension(ex, true)

module MySketchupPlugins
  module AngleDialTool

    class AngleDialTool
      def initialize
        @radius = 50                    # 圆盘半径（屏幕坐标单位）
        @relative_angle = 0.0           # 累积旋转角度（单位：度），可超出显示范围
        @start_angle = nil              # 拖动开始时的包装角度（0~360）
        @prev_angle = nil               # 上一次的包装角度，用于累计差值
        @is_active = false              # 工具激活标志
        @view = nil                     # 当前视图引用

        # 摄像机初始参数（固定）
        @initial_cam_eye = nil
        @initial_cam_up = nil
        @initial_cam_target = nil
        @rotation_axis = nil
      end

      # 辅助方法：将角度包装到 -360 ~ +360 范围内（用于显示）
      def wrap_angle(angle)
        while angle > 360
          angle -= 360
        end
        while angle < -360
          angle += 360
        end
        angle
      end

      def activate
        @is_active = true
        @view = Sketchup.active_model.active_view
        Sketchup.set_status_text("【#{PLUGIN_NAME}】在圆盘上按住[鼠标左键]拖动以旋转视图，按[回车键]可输入数值进行视图旋转（正顺、负逆），按[空格键]退出")
        @view.invalidate  # 立即重绘视图，显示圆盘
      end

      def deactivate(view)
        @is_active = false
        Sketchup::set_status_text("")
        view.invalidate
      end

      def onKeyDown(key, repeat, flags, view)
        return unless @is_active
        if key == 32  # 空格键退出
          Sketchup.active_model.select_tool(nil)
        elsif key == 13  # 回车键确认输入
          input = UI.inputbox(["输入旋转角度 (正数顺时针, 负数逆时针):"], [""], "旋转角度")
          if input && input[0] != ""
            begin
              angle = input[0].to_f
              # 如果没有拖动记录，先设置初始摄像机参数
              if @initial_cam_eye.nil?
                @initial_cam_eye = view.camera.eye.clone
                @initial_cam_up  = view.camera.up.clone
                @initial_cam_target = view.camera.target.clone
                @rotation_axis = view.camera.direction.clone
              end
              @relative_angle += angle
              @relative_angle = wrap_angle(@relative_angle)
              update_view(view)
              view.invalidate
              # 旋转完成后，将当前视图状态设为新的基准，重置圆盘显示为 0
              @initial_cam_eye = view.camera.eye.clone
              @initial_cam_up  = view.camera.up.clone
              @initial_cam_target = view.camera.target.clone
              @rotation_axis = view.camera.direction.clone
              @relative_angle = 0.0
              view.invalidate
            rescue => e
              UI.messagebox("无效的输入，请输入数字！")
            end
          end
        end
      end

      # 返回圆盘中心（始终为视图中心）
      def dial_center(view)
        [view.vpwidth / 2, view.vpheight / 2]
      end

      # 计算屏幕点 pt 相对于 center 的极角（单位：度，范围 0~360）
      def calc_angle(pt, center)
        dx = pt[0] - center[0]
        dy = pt[1] - center[1]
        angle = Math.atan2(dy, dx) * 180.0 / Math::PI
        angle += 360 if angle < 0
        angle
      end

      # 鼠标左键按下时，记录起始角度和摄像机参数
      def onLButtonDown(flags, x, y, view)
        center = dial_center(view)
        distance = Math.sqrt((x - center[0])**2 + (y - center[1])**2)
        if distance <= @radius
          @start_angle = calc_angle([x, y], center)
          @prev_angle = @start_angle
          @relative_angle = 0.0
          @initial_cam_eye = view.camera.eye.clone
          @initial_cam_up  = view.camera.up.clone
          @initial_cam_target = view.camera.target.clone
          @rotation_axis = view.camera.direction.clone
        end
      end

      # 鼠标移动时累计旋转差值（无级累积），更新视图
      def onMouseMove(flags, x, y, view)
        if @start_angle
          center = dial_center(view)
          current_angle = calc_angle([x, y], center)
          delta = current_angle - @prev_angle
          # 修正跨越 0/360 的跳变
          if delta > 180
            delta -= 360
          elsif delta < -180
            delta += 360
          end
          @relative_angle += delta
          @relative_angle = wrap_angle(@relative_angle)
          @prev_angle = current_angle
          update_view(view)
          view.invalidate
        end
      end

      def onLButtonUp(flags, x, y, view)
        @start_angle = nil
        view.invalidate
      end

      def onViewChanged(view)
        view.invalidate if @is_active
      end

      # 绘制圆盘、高亮、刻度及中间角度文本（居中）
      def draw(view)
        center = dial_center(view)

        # 绘制半透明蓝色圆盘底
        bg_color = Sketchup::Color.new(0, 0, 255, 100)
        view.drawing_color = bg_color
        pts = circle_points(center, @radius)
        view.draw2d(GL_POLYGON, pts)

        # 绘制圆盘边界
        view.drawing_color = Sketchup::Color.new(0, 0, 255)
        view.draw2d(GL_LINE_LOOP, pts)

        # 绘制刻度线与刻度标签（固定0°, 30°, …, 330°）
        tick_inner = @radius - 5
        tick_outer = @radius + 10
        (0...360).step(30) do |angle|
          p1 = point_on_circle(center, tick_inner, angle)
          p2 = point_on_circle(center, tick_outer, angle)
          view.draw2d(GL_LINES, [p1, p2])
          tick_label = "#{angle}°"
          raw_pos = point_on_circle(center, tick_outer + 10, angle)
          offset_x = (tick_label.length * 6) / 2.0
          label_pos = [raw_pos[0] - offset_x, raw_pos[1] - 8]
          view.drawing_color = Sketchup::Color.new(0, 0, 255)
          view.draw_text(label_pos, tick_label)
        end

        # 显示用角度：限制在 -360 ~ +360 范围内
        display_angle = wrap_angle(@relative_angle)
        arc_pts = arc_segment_points(center, @radius, 0, display_angle)
        if arc_pts.size > 1
          view.drawing_color = Sketchup::Color.new(255, 0, 0)
          view.draw2d(GL_LINE_STRIP, arc_pts)
        end

        # 绘制中间角度文本，居中对齐
        text = "#{display_angle.to_i}°"
        text_width = text.length * 6   # 每字符约6像素
        text_height = 16               # 估计文本高度16像素
        text_pos = [center[0] - text_width / 2.0, center[1] - text_height / 2.0]
        view.drawing_color = Sketchup::Color.new(0, 255, 0)
        view.draw_text(text_pos, text)

        # 绘制旋转轴（穿过摄像机目标）
        if @initial_cam_target && @rotation_axis
          axis_origin = @initial_cam_target
          axis_dir = @rotation_axis
          pt1 = axis_origin.offset(axis_dir, -100)
          pt2 = axis_origin.offset(axis_dir, 100)
          view.drawing_color = Sketchup::Color.new(0, 0, 0)
          view.draw(GL_LINES, [pt1, pt2])
        end
      end

      # 返回圆盘边界点数组（用于2D绘制）
      def circle_points(center, radius, segments = 72)
        pts = []
        0.upto(segments) do |i|
          theta = 2 * Math::PI * i / segments
          pts << [center[0] + radius * Math.cos(theta), center[1] + radius * Math.sin(theta)]
        end
        pts
      end

      # 返回高亮弧的点数组，从 start_angle 到 end_angle
      def arc_segment_points(center, radius, start_angle, end_angle, segments = 36)
        pts = []
        delta = (end_angle - start_angle) / segments.to_f
        0.upto(segments) do |i|
          angle = start_angle + delta * i
          pts << point_on_circle(center, radius, angle)
        end
        pts
      end

      # 返回圆周上指定角度的点（屏幕坐标）
      def point_on_circle(center, radius, angle_deg)
        rad = angle_deg * Math::PI / 180.0
        [center[0] + radius * Math.cos(rad), center[1] + radius * Math.sin(rad)]
      end

      # 根据累积旋转角度更新视图（摄像机旋转），旋转变换基于全量角度
      def update_view(view)
        return unless @initial_cam_eye && @initial_cam_target && @rotation_axis
        rad = -(@relative_angle * Math::PI / 180.0)
        trans = Geom::Transformation.rotation(@initial_cam_target, @rotation_axis, rad)
        new_eye = @initial_cam_eye.transform(trans)
        new_up  = @initial_cam_up.transform(trans)
        view.camera.set(new_eye, @initial_cam_target, new_up)
      end
    end

    # 插件注册：添加菜单项和工具栏图标按钮
    unless file_loaded?(__FILE__)
      cmd = UI::Command.new(PLUGIN_NAME) {
        Sketchup.active_model.select_tool(AngleDialTool.new)
      }      
      toolbar = UI::Toolbar.new(PLUGIN_NAME)
      cmd = UI::Command.new(PLUGIN_NAME) { Sketchup.active_model.select_tool(AngleDialTool.new) }
      cmd.small_icon = "T_View_Rotation_Tool/View_Rotation.png"
      cmd.large_icon = "T_View_Rotation_Tool/View_Rotation.png"
      cmd.tooltip = PLUGIN_NAME
      cmd.status_bar_text = PLUGIN_DESCRIPTION
      cmd.menu_text = PLUGIN_NAME
      toolbar.add_item cmd
      toolbar.restore
      UI.menu("Plugins").add_item(cmd)
      file_loaded(__FILE__)
    end

  end
end
